import {UNIT_PAY, IS_PAY, MY_BALANCE, PAY_CHECK_INTERVAL} from "./API";
import md5 from 'blueimp-md5'
import qs from 'qs'
import {Utils} from "@/Utils/Utils";
import Decimal from 'decimal.js';
import {createRequest, defaultRequest, DefaultResponse} from "@/Request/Request";
import {
    BACK_ORDER_ID,
    BackInfo,
    DoUnifiedParams,
    FULL_PAY_METHODS,
    OrderStateCallBack,
    PayChannel,
    PayError,
    PayErrorMessage,
    PaymentParams,
    PayMethod,
    PayType,
    ToPayParams,
    ToPayReturns,
    BalanceReturns,
    WechatPayParams
} from "./Defined";
import {ACCESS_TOKEN, UID} from "@/Const";

export * from './Defined'

const SESSION_KEY = 'dc59cf294f37d237c1f06240568ffe21'

const createRequestForPhp = () => {
    const comp = (a: string, b: string): number => {
        if (a > b) {
            return -1;
        } else if (a == b) return 0;
        else return 1;
    }

    const genSign = (data: Record<string, any>): string => {
        let keyArr: string[] = [];
        for (let i in data) {
            keyArr.push(i);
        }
        keyArr.sort(comp);

        const tempArr: string[] = [];
        for (let i in keyArr) {
            tempArr.push(keyArr[i] + "=" + data[keyArr[i]]);
        }
        return md5(tempArr.join("&") + SESSION_KEY);
    }

    const request = createRequest()

    request.interceptors.request.use((config) => {
        const requestData = {
            ...config.data,
            ts: parseInt((Date.now() / 1000).toString()).toString(),
            [UID]: Utils.getQueryString(UID) || Utils.getCookie(UID),
            [ACCESS_TOKEN]: Utils.getQueryString(ACCESS_TOKEN) || Utils.getCookie(ACCESS_TOKEN)
        }
        return {
            ...config,
            transformRequest: [function (data: any, headers: any) {
                return qs.stringify(requestData)
            }],
            headers: {
                'content-type': 'application/x-www-form-urlencoded',
                Authorization: `Ydl ${genSign(requestData)}`,
                [UID]: Utils.getQueryString(UID) || Utils.getCookie(UID),
                [ACCESS_TOKEN]: Utils.getQueryString(ACCESS_TOKEN) || Utils.getCookie(ACCESS_TOKEN)
            }
        }
    })
    return request
}

const requestForPhp = createRequestForPhp()


const requestForJava = defaultRequest

export class Payment {
    balance = 0;
    limitPayChannels: PayChannel[] = [];
    supportCombination: boolean = true
    constructor(paymentParams?: PaymentParams) {
        this.limitPayChannels = paymentParams?.payChannels ?? Object.keys(PayChannel).map(key => key) as PayChannel[];
        if (typeof paymentParams?.supportCombination === 'boolean') {
            this.supportCombination = typeof paymentParams?.supportCombination === 'boolean' && paymentParams.supportCombination
        }
    }

    //isHandleErr 是否要自定义处理异常
    async getBalance(handleErr: (err: BalanceReturns) =>void): Promise<number|null> {
        try {
            const {data: res} = await requestForPhp.post(MY_BALANCE, {balance: 1})
            if (res.data && res.data.balance) {
                this.balance = Number(res.data.balance)
            }
            if(handleErr && res.msg !== 'success'){
                handleErr && handleErr(res)
                return null
            }
            return this.balance
        } catch (err) {
            handleErr && handleErr({success: false})
            return !!handleErr ? null : 0
        }
    }
    getPayMethodList (totalAmount:number) {
        return this.supportCombination ? this.getPayMethodListForCombination(totalAmount) : this.getPayMethodListForNotCombination(totalAmount)
    }

    private getPayMethodListForNotCombination(totalAmount: number): PayMethod[] {
        return FULL_PAY_METHODS.filter(({value}) => {
            switch (value) {
                case PayType.BALANCE:
                    return this.balance > 0
                case PayType.WECHAT:
                    return !Utils.isAlipay()
                case PayType.ALIPAY:
                    return !Utils.isWechat()
                case PayType.HUABEI:
                    return !Utils.isWechat()
            }
        }).map(item => {
            switch (item.value) {
                case PayType.BALANCE:
                    return {
                        ...item,
                        label: `余额<span class="balance">（可用¥${this.balance}）</span>`,
                        disabled:  this.supportCombination ? (this.balance > 0 && this.balance < totalAmount) : true
                    }
            }
            return item
        })
    }
    private getPayMethodListForCombination(totalAmount: number): PayMethod[] {
        return FULL_PAY_METHODS.filter(({value}) => {
            switch (value) {
                case PayType.BALANCE:
                    return this.balance > 0
                case PayType.WECHAT:
                    return this.balance === 0 && !Utils.isAlipay()
                case PayType.WECHAT_BALANCE:
                    return this.balance > 0 && this.balance < totalAmount && !Utils.isAlipay()
                case PayType.ALIPAY:
                    return this.balance === 0 && !Utils.isWechat()
                case PayType.ALIPAY_BALANCE:
                    return this.balance > 0 && this.balance < totalAmount && !Utils.isWechat()
                case PayType.HUABEI:
                    return this.balance === 0 && !Utils.isWechat()
                case PayType.HUABEI_BALANCE:
                    return this.balance > 0 && this.balance < totalAmount && !Utils.isWechat()
            }
        }).map(item => {
            const {payBalance} = this.computeAmountForCombination(totalAmount, this.balance)
            switch (item.value) {
                case PayType.BALANCE:
                    return {
                        ...item,
                        label: `余额<span class="balance">（可用¥${this.balance}）</span>`,
                        disabled: this.balance > 0 && this.balance < totalAmount
                    }
            }
            return item
        })
    }

    computeAmount(totalAmount: number, payType: PayType): { payAmount: number, payBalance: number } {
        let payBalance = 0, payAmount = 0
        if(payType == PayType.BALANCE){
            payBalance = totalAmount;
            payAmount = 0;
        }
        else{
            payBalance = 0;
            payAmount = totalAmount;
        }
        return {payAmount, payBalance}
    }

    computeAmountForCombination(totalAmount: number, balance: number = 0): { payAmount: number, payBalance: number } {
        let payBalance = 0, payAmount = 0
        if (balance === 0) {
            payAmount = totalAmount
        } else if (balance > 0 && balance < totalAmount) {
            payBalance = balance;
            const x = new Decimal(totalAmount);
            const y = new Decimal(balance); 
            payAmount = Number(x.minus(y));
        } else {
            payBalance = totalAmount
        }
        return {payAmount, payBalance}
    } 

    async toPay(params: ToPayParams): Promise<ToPayReturns> {
        const {totalAmount, payType, returnUrl, orderId, subOrderIds, redirectUrl} = params
        let quitUrl = params.quitUrl
        // validate start >>>
        switch (payType) {
            case PayType.BALANCE:
                if (this.balance < totalAmount) {
                    return {errorMessage: PayErrorMessage.BALANCE_EMPTY, success: false, errorType: PayError.VALIDATE}
                }
                break
            case PayType.WECHAT_BALANCE:
            case PayType.WECHAT:
                if (this.limitPayChannels.includes(PayChannel.WX_MWEB) && this.limitPayChannels.includes(PayChannel.WX_JSAPI)) {
                    if (!redirectUrl) {
                        return {
                            errorMessage: PayErrorMessage.WECHAT_REDIRECT_URL_MISSING,
                            success: false,
                            errorType: PayError.VALIDATE
                        }
                    }
                } else {
                    return {
                        errorMessage: PayErrorMessage.WECHAT_PAY_CHANNEL_MISSING,
                        success: false,
                        errorType: PayError.VALIDATE
                    }
                }
                break
            case PayType.ALIPAY:
            case PayType.ALIPAY_BALANCE:
            case PayType.HUABEI:
            case PayType.HUABEI_BALANCE:
            case PayType.HUABEI_STAGE:
            case PayType.HUABEI_STAGE_BALANCE:
                if (this.limitPayChannels.includes(PayChannel.ALI_WAP)) {
                    if (!quitUrl) {
                        return {
                            errorMessage: PayErrorMessage.ALIPAY_QUIT_URL_MISSING,
                            success: false,
                            errorType: PayError.VALIDATE
                        }
                    } else if (!returnUrl) {
                        return {
                            errorMessage: PayErrorMessage.ALIPAY_RETURN_URL_MISSING,
                            success: false,
                            errorType: PayError.VALIDATE
                        }
                    }
                } else {
                    return {
                        errorMessage: PayErrorMessage.ALIPAY_PAY_CHANNEL_MISSING,
                        success: false,
                        errorType: PayError.VALIDATE
                    }
                }
                break
        }
        // validate end <<<
        // hack, 后端微信redirectUrl用的quitUrl
        const payChannel = this.getPayChannel(payType)
        if (payChannel === PayChannel.WX_MWEB || payChannel === PayChannel.WX_JSAPI) {
            quitUrl = redirectUrl
        }
        const { payAmount, payBalance } = this.supportCombination ? this.computeAmountForCombination(totalAmount, this.balance) : this.computeAmount(totalAmount, payType);
        const doUnifiedParams: DoUnifiedParams = {
            returnUrl,
            quitUrl,
            orderId,
            payType,
            payAmount, 
            payBalance,
            subOrderIds
        }
        if (payChannel !== null) {
            doUnifiedParams.payChannel = payChannel
        }
        const res = await requestForJava.post<DoUnifiedParams, DefaultResponse>(UNIT_PAY, doUnifiedParams)
        if (res.code !== '200') return {errorMessage: res.msg as any, success: false, errorType: PayError.BACKEND}
        if (params.payType === PayType.BALANCE && res.data) {
            return {success: res.data.result, errorType: PayError.BALANCE}
        } else if (payChannel === PayChannel.WX_JSAPI) {
            const isPaySucc = await this.wechatPayJSSDK(res.data.content)
            return {
                errorMessage: isPaySucc ? undefined : PayErrorMessage.WECHAT_JSSDK_PAY_ERROR,
                success: isPaySucc,
                errorType: PayError.WECHAT_JSSDK
            }
        } else if (payChannel === PayChannel.WX_MWEB) {
            return {
                payUrl: res.data.content.link,
                success: false,
                errorType: PayError.WECHAT_H5_BREAK,
                errorMessage: PayErrorMessage.WECHAT_H5_PAY_BREAK
            }
        } else if (payChannel === PayChannel.ALI_WAP) {
            return {
                payUrl: res.data.content.link,
                success: false,
                errorType: PayError.ALIPAY_H5_BREAK,
                errorMessage: PayErrorMessage.ALIPAY_H5_PAY_BREAK
            }
        }
        return {success: false, errorType: PayError.UNKNOWN}
    }

    private getPayChannel(payType: PayType): PayChannel | null {
        switch (payType) {
            case PayType.WECHAT_BALANCE:
            case PayType.WECHAT:
                return Utils.isWechat() ? PayChannel.WX_JSAPI : PayChannel.WX_MWEB
            case PayType.ALIPAY:
            case PayType.ALIPAY_BALANCE:
            case PayType.HUABEI:
            case PayType.HUABEI_BALANCE:
            case PayType.HUABEI_STAGE:
            case PayType.HUABEI_STAGE_BALANCE:
                return PayChannel.ALI_WAP
        }
        return null
    }

    private wechatPayJSSDK(wechatPayParams: WechatPayParams): Promise<boolean> {
        const {appId, timeStamp, nonceStr, signType, paySign, package: packageAlias} = wechatPayParams;

        return new Promise((resolve) => {
            window.WeixinJSBridge.invoke(
                "getBrandWCPayRequest",
                {
                    appId: appId, //公众号名称，由商户传入
                    timeStamp: timeStamp, //时间戳，自1970年以来的秒数
                    nonceStr: nonceStr, //随机串
                    package: packageAlias,
                    signType: signType, //微信签名方式：
                    paySign: paySign //微信签名
                },
                (res: any) => {
                    if (res.err_msg === "get_brand_wcpay_request:ok") {
                        resolve(true)
                    } else {
                        resolve(false)
                        console.error(res)
                    }
                }
            );
        })
    }

    static async loopValidateOrderState(count = 0, callBack: OrderStateCallBack, pollTime = 3000): Promise<void> {
        const [, queryString] = window.location.href.split('?')
        const query = queryString ? qs.parse(queryString, {ignoreQueryPrefix: true}) : {}
        const returnParams: BackInfo = {orderId: Number(query[BACK_ORDER_ID])}
        let timer: any = null;
        const checkLoop = () => {
            if (--count > 0) {
                timer = setTimeout(() => {
                    this.loopValidateOrderState(count, callBack, pollTime)
                }, pollTime)
            } else {
                callBack({isPay: false, ...returnParams, isLimit:true})
                timer && clearTimeout(timer)
            }
        }
        if (query[BACK_ORDER_ID]) {
            try {
                const res = await requestForJava.post<Record<string, string>, DefaultResponse>(IS_PAY, {
                    orderId: query[BACK_ORDER_ID]
                })
                if (res.code === '200' && res.data === true) {
                    callBack({isPay: true, ...returnParams})
                    timer && clearTimeout(timer)
                } else {
                    checkLoop()
                }
            } catch (err) {
                checkLoop()
            }
        }
    }

    /**
     * 创建订单查询页面链接， 返回的url不做encode处理
     * @param url
     * @param backOrderId
     */
    static createBackInfoUrl(url: string, backOrderId: number): string {
        const [path, queryString] = url.split('?')
        const query = queryString ? qs.parse(queryString, {ignoreQueryPrefix: true}) : {}
        query[BACK_ORDER_ID] = backOrderId.toString()
        return `${path}?${qs.stringify(query)}`
    }

    static async PayCheckIntervaL(): Promise<number> {
        const res = await requestForJava.get<Record<string, string>, DefaultResponse>(PAY_CHECK_INTERVAL)
        let interval = 3000;
        if(res.code === '200'){
            interval = res.data * 1000;
            return interval
        }
        return interval;
    }
}